(**
 * elaborates abstract syntax tree generated by the parser.
 * This modules does:
 * <dl>
 *   <dt>name resolution</dt>
 *   <dd>translates names in type expressions and module expressions into
 *      FQN of elements they refer to.</dd>
 *   <dt>type annotation</dt>
 *   <dd>matches structures with signatures which constraint them, and
 *      annotates value declarations in structures with their type if they
 *      are specified in the signature.</dd>
 * </dl>
 *
 * @author YAMATODANI Kiyoshi
 * @version $Id: ELABORATOR.sig,v 1.3 2004/11/06 16:15:02 kiyoshiy Exp $
 *)
signature ELABORATOR =
sig

  (***************************************************************************)

  (** long ID *)
  type path = string list

  (***************************************************************************)

  val addExternalType : ((string * path) * ENVSet.ENVSet) -> ENVSet.ENVSet
  val addExternalException : ((string * path) * ENVSet.ENVSet) -> ENVSet.ENVSet
  val addExternalStructure : ((string * path) * ENVSet.ENVSet) -> ENVSet.ENVSet
  val addExternalSignature: ((string * path) * ENVSet.ENVSet) -> ENVSet.ENVSet
  val addExternalFunctor : ((string * path) * ENVSet.ENVSet) -> ENVSet.ENVSet
  val addExternalFunctorSignature :
      ((string * path) * ENVSet.ENVSet) -> ENVSet.ENVSet

  (**
   * do elaboration.
   * @params parameter envSet compileUnits
   * @param parameter global parameter
   * @param envSet initial environment which contains
   * @param compileUnits a list of abstract syntax tree grouped by
   *                compilation unit
   * @return a pair of a new envSet extended by names bound in the given
   *       compilation units and an elaborated abstract syntax tree.
   *
   * @author YAMATODANI Kiyoshi
   * @version 1.0
   *)
  val elaborate :
      DocumentGenerationParameter.parameter ->
      ENVSet.ENVSet ->
      AnnotatedAst.compileUnit list ->
      (ENVSet.ENVSet * ElaboratedAst.compileUnit list)
  
  (***************************************************************************)

end